#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from .quaternions import Quaternion, weighted_average_quaternions

# __init__.py.py
# Created by fzhiheng on 2024/1/16
# Copyright (c) 2024 fzhiheng. All rights reserved.
# 2024/1/16 上午9:34

__all__ = ['Quaternion', 'conversion']
